#pragma once

#ifdef CHATTERINO_HAVE_PLUGINS
#    include "Application.hpp"
#    include "controllers/commands/CommandController.hpp"

#    include <magic_enum.hpp>
#    include <QDir>
#    include <QJsonArray>
#    include <QJsonObject>
#    include <QString>
#    include <semver/semver.hpp>

#    include <set>
#    include <vector>

struct lua_State;

namespace chatterino {

struct PluginMeta {
    // required fields
    QString name;
    QString description;
    std::vector<QString> authors;
    QString license;
    semver::version version;

    // optional
    QString homepage;
    std::vector<QString> tags;

    std::vector<QString> errors;

    bool isValid() const
    {
        return this->errors.empty();
    }

    explicit PluginMeta(const QJsonObject &obj)
        : homepage(obj.value("homepage").toString(""))
    {
        auto nameObj = obj.value("name");
        if (nameObj.isString())
        {
            this->name = nameObj.toString();
        }
        else
        {
            auto type = QString::fromStdString(
                std::string(magic_enum::enum_name(nameObj.type())));
            this->errors.emplace_back(
                QString("name is not a string (its type is %1)").arg(type));
        }

        auto descrObj = obj.value("description");
        if (descrObj.isString())
        {
            this->description = descrObj.toString();
        }
        else
        {
            auto type = QString::fromStdString(
                std::string(magic_enum::enum_name(descrObj.type())));
            this->errors.emplace_back(
                QString("description is not a string (its type is %1)")
                    .arg(type));
        }

        auto authorsObj = obj.value("authors");
        if (authorsObj.isArray())
        {
            auto authorsArr = authorsObj.toArray();
            for (int i = 0; i < authorsArr.size(); i++)
            {
                const auto &t = authorsArr.at(i);
                if (!t.isString())
                {
                    this->errors.push_back(
                        QString(
                            "authors element #%1 is not a string (it is a %2)")
                            .arg(i)
                            .arg(QString::fromStdString(
                                std::string(magic_enum::enum_name(t.type())))));
                    break;
                }
                this->authors.push_back(t.toString());
            }
        }
        else
        {
            auto type = QString::fromStdString(
                std::string(magic_enum::enum_name(authorsObj.type())));
            this->errors.emplace_back(
                QString("authors is not an array (its type is %1)").arg(type));
        }

        auto licenseObj = obj.value("license");
        if (licenseObj.isString())
        {
            this->license = licenseObj.toString();
        }
        else
        {
            auto type = QString::fromStdString(
                std::string(magic_enum::enum_name(licenseObj.type())));
            this->errors.emplace_back(
                QString("license is not a string (its type is %1)").arg(type));
        }

        auto verObj = obj.value("version");
        if (verObj.isString())
        {
            auto v =
                semver::from_string_noexcept(verObj.toString().toStdString());
            if (v.has_value())
            {
                this->version = v.value();
            }
            else
            {
                this->errors.emplace_back(
                    "unable to parse version (use semver)");
                this->version = semver::version(0, 0, 0);
            }
        }
        else
        {
            auto type = QString::fromStdString(
                std::string(magic_enum::enum_name(verObj.type())));
            this->errors.emplace_back(
                QString("version is not a string (its type is %1)").arg(type));
            this->version = semver::version(0, 0, 0);
        }
        auto tagsObj = obj.value("tags");
        if (!tagsObj.isUndefined())
        {
            if (!tagsObj.isArray())
            {
                auto type = QString::fromStdString(
                    std::string(magic_enum::enum_name(licenseObj.type())));
                this->errors.emplace_back(
                    QString("tags is not an array (its type is %1)").arg(type));
                return;
            }

            auto tagsArr = tagsObj.toArray();
            for (int i = 0; i < tagsArr.size(); i++)
            {
                const auto &t = tagsArr.at(i);
                if (!t.isString())
                {
                    this->errors.push_back(
                        QString(
                            "tags element #%1 is not a string (its type is %2)")
                            .arg(i)
                            .arg(QString::fromStdString(
                                std::string(magic_enum::enum_name(t.type())))));
                    return;
                }
                this->tags.push_back(t.toString());
            }
        }
    }
};

class Plugin
{
public:
    QString codename;
    PluginMeta meta;
    bool isDupeName{};

    Plugin(QString codename, lua_State *state, PluginMeta meta,
           const QDir &loadDirectory)
        : codename(std::move(codename))
        , meta(std::move(meta))
        , loadDirectory_(loadDirectory)
        , state_(state)
    {
    }

    ~Plugin();

    /**
     * @brief Perform all necessary tasks to bind a command name to this plugin
     * @param name name of the command to create
     * @param functionName name of the function that should be called when the command is executed
     * @return true if addition succeeded, false otherwise (for example because the command name is already taken)
     */
    bool registerCommand(const QString &name, const QString &functionName);

    /**
     * @brief Get names of all commands belonging to this plugin
     */
    std::set<QString> listRegisteredCommands();

    const QDir &loadDirectory() const
    {
        return this->loadDirectory_;
    }

private:
    QDir loadDirectory_;
    lua_State *state_;

    // maps command name -> function name
    std::map<QString, QString> ownedCommands;

    friend class PluginController;
};
}  // namespace chatterino
#endif
